import pyglet
import pyglet.graphics
from pyglet.sprite import Sprite
from  pyglet.image import SolidColorImagePattern
from pyglet.graphics import OrderedGroup
from pyglet.text import Label
from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED

border = SolidColorImagePattern((0,0,0,	255))
background = SolidColorImagePattern((255,255,255,255))

class WindowManager(object):
	"""Manages windows!"""

	def __init__(self, start = (0,0)):
		self.windows = []
		self.start = start
		self.top = None
		self.dragging = False
		self.over = None
		self.selected = None

	def add(self, window):
		self.windows.append(window)
		self.windows[-1].set_position(self.start)
	
	def set_to_top(self, win):
		if not (self.top == win):
			self.windows.remove(win)
			self.windows.append(win)
			self.top = win
	
	def on_draw(self):
		for window in self.windows:
			window.draw()

	def on_mouse_press(self, x, y,*args, **kwargs):
		win = self.top_window_at((x,y))
		if win:
			self.set_to_top(win)
			it = win.click(x, y, *args, **kwargs)
			if it:
				if self.selected:
					self.selected.un_select()
					self.selected = None
				self.selected = it
				self.selected.select()
			else:
				if self.selected:
					self.selected.un_select()
					self.selected = None
			return EVENT_HANDLED
		else:
			if self.selected:
				self.selected.un_select()
				self.selected = None
			return EVENT_UNHANDLED

	def on_mouse_drag(self, x, y, dx, dy, buttons, *args):
		if self.top:
			if self.top.dragging:
				return self.top.move(dx,dy)
			else:
				return False

	def on_mouse_release(self, x, y, *args):
		win = self.top_window_at((x,y))
		if win:
			return win.un_click(x,y,args)
		else:
			return False

	def on_mouse_motion(self,x,y,dx,dy):
		p = self.top_comp_at((x,y))
		if p and p.overable:
			if self.over:
				if self.over == p:
					return
				else:
					self.over.unset_over()
					self.over = p
					self.over.set_over()
			else:
				self.over = p
				self.over.set_over()
		else:
			if self.over:
				self.over.unset_over()
				self.over = None



	def top_window_at(self, pos):
		for window in reversed(self.windows):
			if window.within(pos):
				return window

		return False

	def top_comp_at(self, pos):
		p = self.top_window_at(pos)
		if p:
			return p.comp_at(pos)
		else:
			return False

	def adjust_point(self, box, point):
		return point[0] + box[0], point[1] + box[1]

	def on_update(self):
		for win in self.windows:
			if win.closed: self.windows.remove(win)

class Component(object):
	"""
	A basic component
	needs to:
	-have a draw method
	-have a position attribute
	-have set_posistion and move functions
	-get_box and within functions for collision detection
	-base 'empty' functions for the possible handlers (click, un_click, key_press, etc)
	"""
	def __init__(self, pos = (0,0), size = (190,20), *args, **kwargs):
		self.pos = pos
		self.size = size
		self.batch = pyglet.graphics.Batch()
		self.overable = False
		self.selectable = False

	def get_box(self):
		return self.pos[0], self.pos[1], self.pos[0]+self.size[0], self.pos[1] + self.size[1]

	def within(self, pos):
		x=pos[0]
		y=pos[1]
		box = self.get_box()
		if (x > box[0]) and (x < box[2]):
			if (y > box[1]) and (y < box[3]):
				return True

	def set_position(self, pos):
		self.pos = pos

	def move(self, dx, dy):
		newPos = self.pos[0]+dx, self.pos[1]+dy
		self.set_position(newPos)
		return True

	def click(self, *args, **kwargs):
		pass

	def un_click(self, *args):
		pass

	def draw(self, *args):
		self.batch.draw()

class LabelWidget(Component):
	""" A widget for a label"""
	def __init__(self, text = "Default",*args, **kwargs):
		super(LabelWidget, self).__init__(*args, **kwargs)
		self.color = (0,0,0,255)
		self.label = Label(text, font_size=10, color=self.color, batch = self.batch, anchor_y = 'top', bold = True)
		self.size = self.label.content_width, self.label.content_height

	def set_position(self, pos):
		super(LabelWidget, self).set_position(pos)
		self.label.x, self.label.y = pos

	def get_box(self):
		return self.pos[0], self.pos[1]-self.label.content_height, self.pos[0]+self.size[0], self.pos[1]


class ButtonWidget(LabelWidget):
	def __init__(self, text = 'Default', pos = (0,0), func = None):
		super(ButtonWidget, self).__init__(text, pos)
		self.over_color = (255,0,0,255)
		self.over = False
		self.overable = True
		self.func = func

	def set_over(self):
		self.label.color = self.over_color

	def unset_over(self):
		self.label.color = self.color

	def click(self, *args, **kwargs):
		if self.func:
			self.func()
		else:
			return False

class FieldWidget(Component):
	"""A widget that holds a selectable and editable 'field'"""
	def __init__(self,default='default', *args, **kwargs):
		super(FieldWidget, self).__init__(*args, **kwargs)
		document = pyglet.text.document.UnformattedDocument(default)
		self.field = pyglet.text.layout.IncrementalTextLayout(document, self.size[0], self.size[1], batch = self.batch)
		self.field.anchor_y = "top"
		self.caret = pyglet.text.caret.Caret(self.field)
		self.content_width = self.field.width
		self.content_height = self.field.content_height
		self.caret_mouse = self.caret.on_mouse_press
		self.caret.on_mouse_press = None
		self.caret.visible = False
		self.selectable = True
		self.selected = False

	def set_position(self, pos):
		super(FieldWidget, self).set_position(pos)
		self.field.x, self.field.y = pos

	def click(self, *args, **kwargs):
		super(FieldWidget, self).click(*args)
		self.caret_mouse(*args, **kwargs)

	def select(self):
		self.selected = True
		self.caret.visible = True
		for window in pyglet.app.windows:
			window.push_handlers(self.caret)
	def un_select(self):
		self.selected = False
		self.caret.visible = False
		for window in pyglet.app.windows:
			window.pop_handlers()

	def get_box(self):
		return self.pos[0], self.pos[1]-self.field.content_height, self.pos[0]+self.size[0], self.pos[1]

class InputWidget(Component):
	def __init__(self, name, default='default', *args, **kwargs):
		super(InputWidget, self).__init__(*args, **kwargs)
		self.name = name
		self.label = LabelWidget(text=name)
		self.field = FieldWidget(default=default)
		self.selectable = True
		self.select = self.field.select
		self.un_select = self.field.un_select

	def set_position(self, pos):
		super(InputWidget, self).set_position(pos)
		self.label.set_position(pos)
		self.field.set_position((pos[0]+self.label.label.content_width+2, pos[1]))

	def value(self):
		return self.field.field.document.text

	def get_box(self):
		return self.pos[0], self.pos[1]-self.field.content_height, self.pos[0]+self.size[0], self.pos[1]

	def draw(self):
		super(InputWidget, self).draw()
		self.label.draw()
		self.field.draw()

	def click(self, *args, **kwargs):
		super(InputWidget, self).click(*args)
		if self.field.within((args[0],args[1])):
			self.field.caret_mouse(*args, **kwargs)

class WindowWidget(Component):
	"""
	TODO:
		The a window holds a single widget.
		-The widget need to be set to the window's batch.
		-

		The window should have a draw or on_draw function.
		-need to test removing an object as a handler from an event dispatch\
			stack. If it works like I want, I can simply remove and push window 
			handlers to handle window draw order.
	"""
	def __init__(self, title = 'default', size = (200,300), group = None, batch = None, pos = (0,0)):
		super(WindowWidget, self).__init__(pos = pos, size = size)
		self.border = Sprite(border.create_image(size[0], size[1]), batch = self.batch, group=pyglet.graphics.OrderedGroup(1))
		self.background = Sprite(background.create_image(size[0]-10, size[1]-20), batch=self.batch, group=pyglet.graphics.OrderedGroup(2))
		self.label = Label(title, font_size=10, color=(255,255,255,255), batch = self.batch, group=pyglet.graphics.OrderedGroup(3), anchor_y = 'top', bold = True)
		self.content = []
		self.dragging = False
		self.closed = False

	def set_position(self, pos):
		super(WindowWidget, self).set_position(pos)
		self.update()

	def click(self, x, y, *args, **kwargs):
		super(WindowWidget, self).click(x,y,args)
		if y > self.pos[1] + self.size[1]-15:
			self.dragging = True
		elif x > self.pos[0] and x < self.pos[0]+5:
			pass
		elif x < self.pos[0]+self.size[0] and x > self.pos[0]+self.size[0]-5:
			pass
		elif y > self.pos[1] and y < self.pos[1]+5:
			pass
		else:
			for item in self.content:
				if item.within((x,y)):
					newargs = [x,y]
					for arg in args:
						newargs.append(arg)
					item.click(*newargs, **kwargs)
					if item.selectable:
						return item
		return False

	def un_click(self, x, y, *args):
		super(WindowWidget, self).un_click(x,y,args)
		if self.dragging: self.dragging = False

	def draw(self):
		super(WindowWidget, self).draw()
		for item in self.content:
			item.draw()

	def update(self):
		self.border.set_position(*self.pos)
		self.background.set_position(self.pos[0]+5, self.pos[1]+5)
		self.label.x, self.label.y = self.pos[0]+5, self.pos[1]+self.size[1]
		j=15
		for item in self.content:
			item.set_position((self.pos[0]+5, self.pos[1]+self.size[1]-j))
			j+=item.size[1]

	def add_component(self, comp):
		self.content.append(comp)
		self.update()

	def on_mouse_motion(self, x, y):
		pass

	def comp_at(self, pos):
		for item in self.content:
			if item.within(pos):
				return item
		return False

class FormWidget(WindowWidget):
	"""A window that holds fields that are compiled on submission"""
	def __init__(self, func, *args, **kwargs):
		super(FormWidget, self).__init__(*args, **kwargs)
		self.func = func
		self.submit_button = ButtonWidget(text="Submit", func = self.submit)
		self.cancel_button = ButtonWidget(text="Cancel", func = self.cancel)
		#self.submit_button.label.anchor_y = "bottom"
		self.fields = []

	def cancel(self):
		self.closed = True
		
	def submit(self):
		dic = {}
		for field in self.fields:
			dic[field.name] = field.value()

		self.func(**dic)

	def add_field(self, field):
		self.content.append(field)
		self.fields.append(field)

	def set_position(self, pos):
		super(FormWidget, self).set_position(pos)
		self.submit_button.set_position((pos[0]+5, pos[1]+20))

	def draw(self):
		super(FormWidget, self).draw()
		self.submit_button.draw()
		#self.cancel_button.draw()
		
	def click(self, x, y, *args, **kwargs):
		if self.submit_button.within((x,y)):
			self.submit_button.click(x, y, *args)
		else:
			return super(FormWidget, self).click(x,y,*args)

