'''
Created on Apr 2, 2010

@author: thebwt
'''
import pyglet
import lib.pygletreactor
lib.pygletreactor.install()
from twisted.internet import reactor

from wengine.modes import BaseMode

class Game(object):
	'''
	The game mode handler

	A Game object starts the window and registers the events needed
	for a bare bones application. Wengine is state, or 'mode' based
	meaning that each mode is isolated completly. See the mode 
	documentation for info on writing custom modes.

	public methods:
	run - runs the main pyglet app loop
	mode_change - changes the game mode, don't use this directly.

	Event types:
	mode_change - is used to change the mode that the Game object is
	in. It takes only a BaseMode (or subclass of) object as a 
	parameter. Implement this to cleanly exit a mode before 
	it is forgotten.
	
	on_update - should not be dispatched manually, it will be done 
	by the pyglet clock automatically. Use an on_update
	function in your states to regularly update modes. Note that this
	is for updates, not drawing; Use it only for things the need regular 
	updates.
	'''
	def __init__(self, starting_mode=BaseMode(), updates_per_second=30):
		'''
		starting_mode - takes a BaseMode object (or subclass of)
		to use as the initial mode.  
		(default wengine.views.base.BaseMode())

		updates_per_second - Takes an Int Sets the rate at which 
		'on_update' is called. (default 30)
		'''
		#Configure the window dispatcher.
		self.win = pyglet.window.Window(caption = "Wengine Base", width = 800, height = 600)
		self.win.register_event_type('mode_change')
		self.win.register_event_type('on_update')
		self.win.push_handlers(self)
		pyglet.clock.schedule_interval(lambda x: self.win.dispatch_event('on_update'), 1.0/updates_per_second)
		
		#initial mode settings go here.
		self.mode = None
		self.win.dispatch_event('mode_change', starting_mode)


	def run(self):
		"""Run the Game object event loop"""
		reactor.run()
	
	def force_mode(self, mode):
		self.win.dispatch_event('mode_change', mode)

	def mode_change(self, mode):
		"""
		Change the Game's mode

		This should not be invoked directly, have the window
		dispatch a mode_change event.
		"""
		if self.mode: 
			self.win.pop_handlers()
			self.mode.close()
		self.mode = mode
		self.mode.start()
		self.win.push_handlers(self.mode)

	def on_close(self, *arg):
		"""
		Quit the Game

		this is the correct way to shut down, app.quit()
		doesn't work with the pyglet-twisted loop
		"""
		reactor.callFromThread(reactor.stop)
		return True
